import 'package:analyzer/dart/constant/value.dart';
import 'package:analyzer/dart/element/element.dart';
import 'package:analyzer/dart/element/nullability_suffix.dart';
import 'package:analyzer/dart/element/type.dart';
import 'package:analyzer/dart/element/visitor.dart';
import 'package:build/build.dart';
import 'package:controller/controller.dart';
import 'package:controller_generator/src/util.dart';
import 'package:source_gen/source_gen.dart';

const httpRequestMeta = {
  'Get',
  'Put',
  'Post',
  'Delete',
  'Options',
  'Patch',
  'Head',
  'HttpRequest',
};

typedef TypeCheck = bool Function(DartType);

const coreValidators = <TypeCheck, String>{
  isNum: 'number',
  isInt: 'integer',
  isDouble: 'decimal',
};

bool isNum(DartType t) => t.isDartCoreNum;

bool isInt(DartType t) => t.isDartCoreInt;

bool isDouble(DartType t) => t.isDartCoreDouble;

class DispatcherBuilder extends GeneratorForAnnotation<Controller> {
  @override
  String generateForAnnotatedElement(
      Element element, ConstantReader annotation, BuildStep buildStep) {
    if (element.kind != ElementKind.CLASS) {
      throw 'ERROR: @Constructor can only be used on a class, found on $element';
    }
    var typeName = element.name;
    var matcherBuilder = MatcherBuilder();
    element.visitChildren(matcherBuilder);

    if (matcherBuilder.matchers.isEmpty) {
      return '';
    }
    return '''
    class ${typeName}\$DispatcherBuilder extends DispatcherBuilder {
      final ${typeName} controller;
      ${matcherBuilder.hasSecurity ? 'final Security security;' : ''}

      ${typeName}\$DispatcherBuilder(this.controller${matcherBuilder.hasSecurity ? ', this.security' : ''});
      
      @override
      AsyncPatternMatcher<Request, Response> appendMatchers(
          AsyncPatternMatcher<Request, Response> requestMatcher) {
        return requestMatcher
        ${matcherBuilder.matchers.join('\n')};
      }
    }
    ''';
  }
}

class MatcherBuilder extends SimpleElementVisitor<void> {
  final List<String> matchers = [];
  bool hasSecurity = false;

  @override
  void visitMethodElement(MethodElement element) {
    var httpRequests = element.metadata.map(_extractHttpRequest).whereNotNull();
    var security = _getMetadataOfType(
        element.metadata,
        'Secured',
        (value) => Secured(value
            .getField('claims')
            .toListValue()
            .map((claim) => claim.toStringValue())
            .toList()));
    if (security != null) {
      hasSecurity = true;
    }

    matchers.add(httpRequests
        .map((httpRequest) =>
            _buildRequestHandler(element, httpRequest, security))
        .join('\n'));
  }

  T /*?*/ _getMetadataOfType<T>(List<ElementAnnotation> metadata, String name,
          T Function(DartObject object) constructor) =>
      metadata.map((meta) {
        var value = meta.computeConstantValue();
        if (value.type.element.library.name == 'controller.meta' &&
            value.type.element.name == name) {
          return constructor(value);
        }
        return null;
      }).firstWhere((element) => element != null, orElse: () => null);

  HttpRequest /*?*/ _extractHttpRequest(ElementAnnotation element) {
    var value = element.computeConstantValue();
    if (value.type.element.library.name == 'controller.meta' &&
        httpRequestMeta.contains(value.type.element.name)) {
      return _constructHttpRequest(value);
    }
    return null;
  }

  HttpRequest _constructHttpRequest(DartObject value) {
    if (value.type.element.name == 'HttpRequest') {
      return HttpRequest(value.getField('method').toStringValue(),
          value.getField('path').toStringValue());
    } else {
      return _constructHttpRequest(value.getField('(super)'));
    }
  }

  String _buildRequestHandler(
      MethodElement element, HttpRequest httpRequest, Secured /*?*/ security) {
    var bindArguments = _bindArguments(element);
    var decodeBody = '';
    if (bindArguments.contains('decodedBody')) {
      decodeBody = 'var decodedBody = jsonDecode(body);';
    }
    var checkSecurity = _buildSecurityCheck(security);
    var validate = _buildValidation(element);
    return '''
    .when2<Map<String, String>, dynamic>(
      matchRequest('${httpRequest.method}', '${httpRequest.path}'),
      (headers, body) async {
        $checkSecurity
        $decodeBody
        $bindArguments
        $validate
        return await controller.${element.name}(${_parameters(element)});
      },
    )
    ''';
  }

  String _parameters(MethodElement element) =>
      element.parameters.map((parameter) {
        var converted = _convertIfNecessary(
            parameter.type, parameter.name, parameter.defaultValueCode);
        return parameter.hasDefaultValue ||
                parameter.type.nullabilitySuffix != NullabilitySuffix.none
            ? converted
            : '${converted}/*!*/';
      }).join(', ');

  String _buildSecurityCheck(Secured /*?*/ security) {
    return security != null
        ? '''
      if (!security.verify(headers, [${security.claims.map((claim) => "'$claim'").join(',')}])) {
        return Response(HttpStatus.unauthorized);
      }
      '''
        : '';
  }

  String _bindArguments(MethodElement element) =>
      '${element.parameters.map(_bindArgument).join()}';

  String _bindArgument(ParameterElement element) => _isBody(element)
      ? 'var ${element.name} = ${element.type.element.name}.fromJson(decodedBody);\n'
      : 'var ${element.name} = ${_isBodyProperty(element) ? "decodedBody['${element.name}']" : "headers['${element.name}']"};\n';

  String _convertIfNecessary(
          DartType type, String argument, String defaultValue) =>
      _applyDefault(
          argument, 'convert<${type.element.name}>($argument)', defaultValue);

  String _applyDefault(
          String argument, String expression, String /*?*/ defaultValue) =>
      argument == expression && defaultValue == null
          ? argument
          : '$argument != null ? $expression : $defaultValue';

  bool _isBody(ParameterElement element) => hasMeta(element, Body);

  bool _isBodyProperty(ParameterElement element) =>
      hasMeta(element, BodyProperty);

  String _buildValidation(MethodElement element) {
    var validators = element.parameters
        .expand((parameter) {
          var validators = <String /*?*/ >[];
          _addEntityValidator(parameter, validators);
          _addRequiredValidator(parameter, validators);
          _addCoreValidators(parameter, validators);
          _addCustomValidators(parameter, validators);
          return validators;
        })
        .where((validator) => validator != null)
        .toList();
    return validators.isNotEmpty
        ? '''
    var errors = collectErrors([${validators.join(', ')}]);
    if (errors.isNotEmpty) {
      return badRequest(errors);
    }
    '''
        : '';
  }

  void _addCustomValidators(
      ParameterElement parameter, List<String> validators) {
    validators.addAll(parameter.metadata
        .where((metadata) =>
            isType(metadata.computeConstantValue().type, PropertyValidator))
        .map((metadata) => 'await ${metadata.toSource().substring(1)}'
            ".validate('${parameter.name}', ${parameter.name})"));
  }

  void _addCoreValidators(ParameterElement parameter, List<String> validators) {
    return coreValidators.forEach((check, validator) {
      if (check(parameter.type)) {
        validators
            .add("$validator.validate('${parameter.name}', ${parameter.name})");
      }
    });
  }

  void _addRequiredValidator(
      ParameterElement parameter, List<String> validators) {
    if (parameter.type.nullabilitySuffix == NullabilitySuffix.none) {
      validators
          .add("required.validate('${parameter.name}', ${parameter.name})");
    }
  }

  void _addEntityValidator(
      ParameterElement parameter, List<String> validators) {
    if (_isValidatable(parameter.type.element)) {
      validators.add(
          'await ${parameter.type.element.name}\$Validator.instance.validateEntity(${parameter.name})');
    }
  }

  bool _isValidatable(Element element) => hasMeta(element, Validatable);
}

extension NotNull<T> on Iterable<T /*?*/ > {
  Iterable<T> whereNotNull() => where((e) => e != null).map((e) => e /*!*/);
}
